// jzLintReport.cpp : Defines the entry point for the console application.
//
// This is designed to be used in a DOS batch file
// It will scan thru the PC-LINT output, and generate
// a report of the errors, sorted by message number
// 
// This makes it easier to scan thru the errors, and
// decide which errors to concentrate on
//
// ie, the PC-LINT tolerance can be set low...
// usually this results in too many error messages,
// however using this method, the "clutter" can be
// organized
// 

#include "stdafx.h"

typedef char *LPSTR;
typedef const LPSTR LPCSTR;
typedef unsigned char BOOL;
#define FALSE ( 0 )
#define TRUE ( 1 )
#define WORD unsigned int
//lint -esym(421, gets) // turn off Warning 421: Caution -- function 'gets(char *)' is considered dangerousWarning 421: Caution -- function 'gets(char *)' is considered dangerous

#define OTHER_TITLE "OTHER - Messages that haven't been defined in a category"

//********************************************************************
//*
//* The main structures used by the program are defined here
//*
//********************************************************************

typedef struct // This is used for each message generated by PC-LINT
{
	int m_nErrNo;
	int m_nCatNo;
	int m_nCatIndex;
	LPSTR m_pLine;
} LINTLINE;

typedef struct // This is used for each file processed by this program
{
public:
	LPSTR m_pFileName;
	int m_nMessageCount;
} FILENAMES;

typedef struct // This is used for each message Category defined
{
	LPSTR m_pCatName;
	int   m_nCount; // the number of messages defined in this category
} DEFINED_CATEGORY;

typedef struct // This is used for each message defined (used to map message number to category name)
{
	int m_nMsgNo;
	int m_nCatNo; // index into CateforyList
} DEFINED_MESSAGE;


//********************************************************************
//*
//* The main array max sizes are defined here
//*
//********************************************************************

#define MAX_FILES ( 1024 * 4 )
#define MAX_LINE ( 100 * 1024)
#define MAX_DEFINED_CATEGORY ( 4 * 1024 )
#define MAX_DEFINED_MESSAGES ( 10 * 1024 ) // list of possible messages that PC-Lint can generate

#define MAX_LINE_LENGTH ( 8 * 4096 ) // the maximum lenght of a single line of text read in from a file
//********************************************************************
//*
//* The main arrays are defined here
//*
//********************************************************************

DEFINED_CATEGORY   DefinedCategoryList[MAX_DEFINED_CATEGORY];
DEFINED_MESSAGE    DefinedMessageList[MAX_DEFINED_MESSAGES];
LINTLINE           LintLine[MAX_LINE];
FILENAMES          FileNames[MAX_FILES];

//********************************************************************
//*
//* The main array counts (currently used bounds) are defined here
//*
//********************************************************************

int nDefinedCategoryCount = 0;
int nDefinedMessageCount = 0; 
int nLintLineCount = 0;
int nFileNamesCount = 0;
int nCatNo_Other; // the category index for "Other" in "DefinedMessageList"

//********************************************************************
//* this routine looks up the message number in "DefinedMessageList"
//* any undefined message is returned as category "OTHER" 
//* The array "DefinedMessageList" is assumed to be sorted, so a binary
//* search is used
//********************************************************************
int LookUpMessageInCategoryList(int nMsgNo, int &nIndex) // returns the index of the message in the "DefinedCategoryList"
{
	int nLeft;
	int nRight;
	int nMid;
	nIndex = -1;
	nLeft = 0;
	nRight = nDefinedMessageCount-1;

	if (nMsgNo == 801)
	{
		int bFlag = TRUE;
	}

	while (nLeft <= nRight)
	{
		nMid = (nLeft + nRight)/2;
		if (nMsgNo < DefinedMessageList[nMid].m_nMsgNo)
		{
			nRight = nMid-1;
		}
		else if (nMsgNo > DefinedMessageList[nMid].m_nMsgNo)
		{
			nLeft = nMid+1;
		}
		else // if (DefinedMessageList[nMid].m_nMsgNo == nMsgNo)
		{
			nIndex = nMid;
			return DefinedMessageList[nMid].m_nCatNo;
		}
	}
	return nCatNo_Other; // not found, return it as the "OTHER" category
}

//********************************************************************
//*
//* The main array qsort funstions are defined here
//*
//********************************************************************

// used to sort the messages by error number
int FuncCmpLintLine(const void *elem1, const void *elem2)
{
	LINTLINE *pa = (LINTLINE*) elem1;
	LINTLINE *pb = (LINTLINE*) elem2;

	// first test the categories
	if (pa->m_nCatNo < pb->m_nCatNo)
		return -1;
	if (pa->m_nCatNo > pb->m_nCatNo)
		return 1;

	// if they are in the same category, 
	// then test the order they were defined
	if (pa->m_nCatIndex < pb->m_nCatIndex)
		return -1;
	if (pa->m_nCatIndex > pb->m_nCatIndex)
		return 1;

	// they are eight the same sort values...
	// or not defined, so in this case,
	// sort by the actual message number

	if (pa->m_nErrNo < pb->m_nErrNo)
		return -1;
	if (pa->m_nErrNo > pb->m_nErrNo)
		return 1;
	return stricmp(pa->m_pLine, pb->m_pLine);
}	

// used to sort the messages by file name
int FuncCmpFileNames(const void *elem1, const void *elem2)
{
	FILENAMES *pa = (FILENAMES*) elem1;
	FILENAMES *pb = (FILENAMES*) elem2;
	return stricmp(pa->m_pFileName, pb->m_pFileName);
}	
// sorts descending by number of messages
int FuncCmpFileNamesByCount(const void *elem1, const void *elem2)
{
	FILENAMES *pa = (FILENAMES*) elem1;
	FILENAMES *pb = (FILENAMES*) elem2;
	if (pa->m_nMessageCount > pb->m_nMessageCount)
		return -1;
	if (pa->m_nMessageCount < pb->m_nMessageCount)
		return 1;
	return 0;
}	
// used to sort the messages defined in "DefinedMessageList"
int FuncCmpLintDefinedMessageList(const void *elem1, const void *elem2)
{
	DEFINED_MESSAGE *pa = (DEFINED_MESSAGE*) elem1;
	DEFINED_MESSAGE *pb = (DEFINED_MESSAGE*) elem2;
	if (pa->m_nMsgNo < pb->m_nMsgNo)
		return -1;
	if (pa->m_nMsgNo > pb->m_nMsgNo)
		return 1;
	return 0;
}	


// messages are detected by scanning for strings with
// the character sequence ") : xxxx nnnn:"
// Here's some examples of messages we're scanning for:
// C:\Projects\WinDNM\COMPUTE.CPP(95) :  Info 1740: pointer member 'CDNMComputeUsingRecipe::m_pCompute' (line 63) not directly freed or zero'ed by destructor
// C:\Projects\WinDNM\COMPUTE.CPP(834) :  Warning 527: Unreachable
// C:\Projects\WinDNM\DBFIELD.CPP(1001) :  Error 1023: Call 'COleDateTime::operator>(COleDateTime, int)' is ambiguous;  candidates: 'COleDateTime::operator>(const COleDateTime &) const (line 1152, file C:\Program Files\Microsoft Visual Studio\VC98\MFC\INCLUDE\afxdisp.h), and operator>(double, int) (location unknown)'
// C:\Projects\WinDNM\Compute.cpp(802) :  Note 944: Left argument for operator '||' always evaluates to False
// --- Module:   C:\Projects\WinDNM\RECIPE.CPP 

#define MODULE_STR "--- Module:"
// assumes called with pFileName is NULL
int ParseLine(LPCSTR s, LPSTR &pFileName)
{
	WORD i;
	WORD nLen = strlen(s);

	if (memicmp(s, MODULE_STR, strlen(MODULE_STR)) == 0)
	{
		LPSTR p = s + strlen(MODULE_STR)+1;
		// skip leading spaces
		while (*p && ((*p == ' ') || (*p == '\t')))
			p++;
		WORD nLen = strlen(p);
		// skip trailing spaces
		while (nLen && ((p[nLen-1] == ' ') || (p[nLen-1] == '\t') || (p[nLen-1] == '\n') || (p[nLen-1] == '\r')))
			nLen--;
		pFileName = (LPSTR) malloc(nLen+1);
		for (i = 0; i < nLen; i++)
			pFileName[i] = p[i];
		pFileName[i] = 0;
		return -1;
	}

	// scan for leading ")"
	for (i = 0; i < nLen; i++)
	{
		if (s[i] == ')')
			break;
	}
	if (s[i] != ')')
		return -1;
	i++; // eat ')'

	// skip past any white space...
	for (; i < nLen; i++)
	{
		if ((s[i] != ' ') && (s[i] != '\t'))
			break;
	}

	if (s[i] != ':')
		return -1;
	i++; // eat ':'

	// skip past any white space...
	for (; i < nLen; i++)
	{
		if ((s[i] != ' ') && (s[i] != '\t'))
			break;
	}

	// skip past message type ("Info", "Warning", "Error")...
	if (memicmp(&s[i], "Info", strlen("Info")) == 0)
	{
		i += strlen("Info");
	}
	else if (memicmp(&s[i], "Warning", strlen("Warning")) == 0)
	{
		i += strlen("Warning");
	}
	else if (memicmp(&s[i], "Error", strlen("Error")) == 0)
	{
		i += strlen("Error");
	}
	else if (memicmp(&s[i], "Note", strlen("Note")) == 0)
	{
		i += strlen("Note");
	}
	else
		return -1;

	// skip past any white space...
	for (; i < nLen; i++)
	{
		if ((s[i] != ' ') && (s[i] != '\t'))
			break;
	}

	// read number...
	int nRet = 0;
	for (; i < nLen; i++)
	{
		if ((s[i] < '0') || (s[i] > '9'))
			break;
		nRet = nRet*10 + (s[i] - '0');
	}

	// save the file name...
	for (nLen = 0; (s[nLen] && (s[nLen] != '(')); nLen++)
		;

	pFileName = (LPSTR) malloc(nLen+1);
	for (i = 0; i < nLen; i++)
	{
		pFileName[i] = s[i];
	}
	pFileName[i] = 0;
	return nRet;
}

// returns the index of the category name in "DefinedCategoryList"
int DefineCategory(LPSTR szCatName, LPSTR szFileName)
{
	int i;
	// First check if the category has already been defined
	// use linear lookup, as the number of categories is usually low
	for (i = 0; i < nDefinedCategoryCount; i++)
	{
		if (stricmp(szCatName, DefinedCategoryList[i].m_pCatName) == 0)
			return i;
	}
	if (nDefinedCategoryCount >= MAX_DEFINED_CATEGORY)
	{
		printf("\nError: Too many message categories defined in file: %s\n", szFileName);
		printf("Error: Category name was: %s\n", szCatName);
		return 0;
	}
	LPSTR pName = (LPSTR) malloc(strlen(szCatName)+1);
	if (!pName)
		return nCatNo_Other;
	strcpy(pName, szCatName);
	DefinedCategoryList[nDefinedCategoryCount].m_pCatName = pName;
	DefinedCategoryList[nDefinedCategoryCount].m_nCount = 0;
	nDefinedCategoryCount++;
	return nDefinedCategoryCount-1;
}

// The file format is a simple text file
// blank lines are ignored
// leading white space is ignored
// comments are ignored (anything after '//')
// the format of a line to start a category is:
//              $$<name> 
// anything else is a message of the form:
// 
//	      Info 1740: pointer member 'CDNMComputeUsingRecipe::m_pCompute' (line 63) not directly freed or zero'ed by destructor
//	      Warning 527: Unreachable
//	      Error 1023: Call 'COleDateTime::operator>(COleDateTime, int)' is ambiguous;  candidates: 'COleDateTime::operator>(const COleDateTime &) const (line 1152, file C:\Program Files\Microsoft Visual Studio\VC98\MFC\INCLUDE\afxdisp.h), and operator>(double, int) (location unknown)'
//	      Note 944: Left argument for operator '||' always evaluates to False
//
// Note that only the message number is used, the other text is allowed for
// easy cut-and-paste, and to be somewhat self-documenting
void ReadInDefinedCategories(LPSTR szFileName)
{
	int i;
	char line[MAX_LINE_LENGTH];
	char NewCatName[MAX_LINE_LENGTH];
	LPSTR pNewCatName; 
	int nCatNo = -1;  // the first category is always "OTHER"

	nDefinedCategoryCount = 0;
	nDefinedMessageCount = 0;

	FILE *f = fopen(szFileName, "rt");
	if (!f)
	{
		printf("\nError: could not open file: %s\n", szFileName);
	}
	else
	{
		while (fgets(line, sizeof(line), f))
		{
			char *pLine = line;
			// trim trailing eof
			for (i = 0; line[i]; i++)
			{
				if ((line[i] == '\n') || (line[i+1] == '\r'))
				{
					line[i] = '\0';
					break;
				}
			}

			// trim comments ('//'), if any
			for (i = 0; line[i] && line[i+1]; i++)
			{
				if ((line[i] == '/') && (line[i+1] == '/'))
				{
					line[i] = '\0';
					break;
				}
			}
			// trim leading spaces
			LPSTR p = line;
			while (*p && ((*p == ' ') || (*p == '\t')))
				p++;
			if (!*p)
				continue; // blank line, ignore it and continue reading lines
			// check if start of a category....
			if (*p == '$')
			{
				if (memicmp(p, "$$", strlen("$$")) != 0)
					continue; // not a valid line, ignore it and continue parsing
				p += strlen("$$");
				while (*p && (*p == ' ') && (*p == '\t')) // skip any white space
					p++;
				memset(NewCatName, 0, sizeof(NewCatName));
				pNewCatName = NewCatName;
				while (*p && (*p != '\n') && (*p != '\r'))
				{
					*pNewCatName = *p;
					pNewCatName++;
					p++;
				}
				// remove any trailing spaces
				for (i = (int) strlen(NewCatName)-1; i >= 0; i--)
				{
					char ch = NewCatName[i];
					if ((ch == ' ') || (ch == '\t'))
						NewCatName[i] = '\0';
					else
						break;
				}
				nCatNo = DefineCategory(NewCatName, szFileName);
				continue; // get next line
			}

			// not a category, then its a message number line
			// skip leading decoration
			while (*p && ((*p < '0') || (*p > '9')))
				p++;
			if (!*p)
				continue; // blank line, ignore it and continue reading in lines
			// parse message number
			int nMsgNo = 0;
			while (*p && ((*p >= '0') && (*p <= '9')))
			{
				nMsgNo = nMsgNo*10 + (*p - '0');
				p++;
			}
			if (nCatNo < 0)
			{
				// there was an error in the definition file,
				// the messages started without a category header...
				// so, create a category...
				nCatNo =  DefineCategory("UNDEFINED CATEGORY", szFileName); 
			}
			DefinedMessageList[nDefinedMessageCount].m_nMsgNo = nMsgNo;
			DefinedMessageList[nDefinedMessageCount].m_nCatNo = nCatNo; 
			DefinedCategoryList[nCatNo].m_nCount++;
			if (nDefinedMessageCount >= MAX_DEFINED_MESSAGES)
			{
				printf("\nError: Too many message numbers defined in file: %s\n", szFileName);
				break;
			}
			nDefinedMessageCount++;
		}
		fclose(f);
	}
	// special case, any undefined message will be in this category
	// Make sure its defined, if the user hasn't already defined it
	nCatNo_Other = DefineCategory(OTHER_TITLE, szFileName); 
	qsort(DefinedMessageList, nDefinedMessageCount, sizeof(DefinedMessageList[0]), FuncCmpLintDefinedMessageList);
}

int main(int argc, char* argv[])
{
	char Line[MAX_LINE_LENGTH];
	int i;
	LPSTR  pFileName = NULL;


	LPSTR pCategoryFileName = argv[1]; 
	if (!pCategoryFileName)
	 pCategoryFileName = "c:\\lint\\MsgCats.txt";

	ReadInDefinedCategories(pCategoryFileName);

	nLintLineCount = 0;
	while (gets(Line))
	{
//		puts(Line); // debug line-by-line output
		int n = ParseLine(Line, pFileName);
		if (n == 831) // skip these messages for the summary: Info 831: Reference cited in prior message
			continue;
		if (n >= 0)
		{
			if (nLintLineCount < MAX_LINE)
			{
				int nCatIndex;
				int nCat = LookUpMessageInCategoryList(n, nCatIndex);
				LintLine[nLintLineCount].m_nErrNo = n;
				LintLine[nLintLineCount].m_pLine = (LPSTR) malloc(strlen(Line)+1);
				strcpy(LintLine[nLintLineCount].m_pLine, Line);
				LintLine[nLintLineCount].m_nCatNo = nCat;
				LintLine[nLintLineCount].m_nCatIndex = nCatIndex;
				nLintLineCount++;
			}
		}
		if (pFileName)
		{
			BOOL bFound = FALSE;
			for (i = nFileNamesCount-1; i >= 0; i--)
			{
				if (stricmp(FileNames[i].m_pFileName, pFileName) == 0)
				{
					bFound = TRUE;
					if (n >= 0)
					{
						FileNames[i].m_nMessageCount++;
					}
				}
			}
			if (bFound)
			{
				free(pFileName);
				pFileName = NULL;
			}
			else
			{
				FileNames[nFileNamesCount].m_pFileName = pFileName;
				pFileName = NULL;
				FileNames[nFileNamesCount].m_nMessageCount = 0;
				nFileNamesCount++;
			}
		}
	}

	qsort(LintLine, nLintLineCount, sizeof(LINTLINE), FuncCmpLintLine);
	qsort(FileNames, nFileNamesCount, sizeof(FILENAMES), FuncCmpFileNames);

	// print report
	puts("\n\n********************************************************\n");
	puts(   "*            Lint Error/Warning messages follow         *\n");
	puts(   "*********************************************************\n\n");
	int nCatNo = -1;

	if (nLintLineCount <= 0)
	{
		printf("********** Finished processing, no messages detected\n");
	}
	for (i = 0; i < nLintLineCount; i++)
	{
		if (LintLine[i].m_pLine)
		{
			if (LintLine[i].m_nCatNo != nCatNo)
			{
				nCatNo++;
				while (nCatNo < LintLine[i].m_nCatNo)
				{
					printf("********** Category: %s\n", DefinedCategoryList[nCatNo].m_pCatName);
					if (DefinedCategoryList[nCatNo].m_nCount == 0)
						printf(" No messages have been defined for this category\n");
					else
						printf(" No messages in this category\n");
					nCatNo++;
				}
				printf("********** Category: %s\n", DefinedCategoryList[nCatNo].m_pCatName);
			}
			puts(LintLine[i].m_pLine);
		}
	}
	// generate trailing messages
	nCatNo++;
	while (nCatNo < nDefinedCategoryCount)
	{
		printf("********** Category: %s\n", DefinedCategoryList[nCatNo].m_pCatName);
		if (DefinedCategoryList[nCatNo].m_nCount == 0)
			printf(" No messages have been defined for this category\n");
		else
			printf(" No messages in this category\n");
		nCatNo++;
	}

	int nTotalFilesWithMessages = 0;
	for (i = 0; i < nFileNamesCount; i++)
	{
		if (FileNames[i].m_nMessageCount)
			nTotalFilesWithMessages++;
	}

	puts("");
	char buf[4096];
	sprintf(buf, "Total Files Processed: %d", nFileNamesCount);
	puts(buf);
	sprintf(buf, "Total Files With Messages: %d", nTotalFilesWithMessages);
	puts(buf);
	sprintf(buf, "Total Messages Processed: %d", nLintLineCount);
	puts(buf);


	if (nFileNamesCount > 1)
	{
		// print list of files with messages
		puts("Processed Files with PC-LINT messages:");
		for (i = 0; i < nFileNamesCount; i++)
		{
			if (FileNames[i].m_nMessageCount)
			{
				sprintf(buf, "    %4d %s", FileNames[i].m_nMessageCount, FileNames[i].m_pFileName);
				puts(buf);
			}
		}
		qsort(FileNames, nFileNamesCount, sizeof(FILENAMES), FuncCmpFileNamesByCount);
		// print list of files with messages
		puts("Processed Files with PC-LINT messages sorted by number of errors:");
		for (i = 0; i < nFileNamesCount; i++)
		{
			if (FileNames[i].m_nMessageCount)
			{
				sprintf(buf, "    %4d %s", FileNames[i].m_nMessageCount, FileNames[i].m_pFileName);
				puts(buf);
			}
		}
	
		qsort(FileNames, nFileNamesCount, sizeof(FILENAMES), FuncCmpFileNames);
		// print list of files with no messages
		puts("Processed Files with NO PC-LINT messages:");
		for (i = 0; i < nFileNamesCount; i++)
		{
			if (FileNames[i].m_nMessageCount == 0)
			{
				sprintf(buf, "    %4d %s", FileNames[i].m_nMessageCount, FileNames[i].m_pFileName);
				puts(buf);
			}
		}
	}

	// free memory
	for (i = 0; i < nLintLineCount; i++)
	{
		if (LintLine[i].m_pLine)
			free(LintLine[i].m_pLine);
	}
	// free memory
	for (i = 0; i < nFileNamesCount; i++)
	{
		if (FileNames[i].m_pFileName)
			free(FileNames[i].m_pFileName);
	}
	return 0;
}

