package com.ddj.ddjplayer;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.Properties;
import java.io.FileInputStream;
import java.io.File;

public class DdjFrame extends JFrame {
  protected PlayerBase pBase = null;
  protected String searchDir = null;
  protected int curPlayListLength = 0;
  static final String MUSIC_DIR_PROPERTY = "musicdir";
  static final String DEFAULT_DIR = "music";
  static final String CONFIG_FILE_NAME = "ddjplay.ini";

  private JPanel refContentPane;

  private JMenuBar playerMenu = new JMenuBar();
  private JMenu menuFile = new JMenu();
  private JMenuItem menuFileExit = new JMenuItem();
  private JMenu menuHelp = new JMenu();
  private JMenuItem menuHelpAbout = new JMenuItem();


  private JPanel listPanel = new JPanel();
  private  JPanel buttonPanel = new JPanel();
  private  JButton prevButton = new JButton();

  JButton playButton = new JButton();
  JButton pauseButton = new JButton();
  JButton stopButton = new JButton();
  JList filenamesList = new JList();
  JButton nextButton = new JButton();

  public DdjFrame() {
    enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    try {
      guiInit();
      logicInit();
      pBase = new PlayerBase();
    }
    catch(Exception e) {
      e.printStackTrace();
    }
  }

  private void guiInit() throws Exception  {
    refContentPane = (JPanel) this.getContentPane();
    refContentPane.setLayout(new BorderLayout());
    this.setSize(new Dimension(400, 300));
    this.setTitle("DDJ Java Music Player");
    menuFile.setText("File");
    menuFileExit.setText("Exit");
    menuFileExit.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuFileExitClicked(e);
      }
    });
    menuHelp.setText("Help");
    menuHelpAbout.setText("About");
    menuHelpAbout.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuHelpAboutClicked(e);
      }
    });

    buttonPanel.setPreferredSize(new Dimension(30, 100));
    buttonPanel.setLayout(new GridLayout());

    addButtonIconText(prevButton, "Prev", "StepBack24.gif");
    prevButton.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(ActionEvent e) {
        prevClicked(e);
      }
    });


    addButtonIconText(playButton, "Play", "Play24.gif");
    playButton.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(ActionEvent e) {
        playClicked(e);
      }
    });


    addButtonIconText(pauseButton, "Pause", "Pause24.gif");
     pauseButton.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(ActionEvent e) {
        pauseClicked(e);
      }
    });

    addButtonIconText(stopButton, "Stop", "Stop24.gif");
     stopButton.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(ActionEvent e) {
        stopClicked(e);
      }
    });


    addButtonIconText(nextButton, "Next", "StepForward24.gif");
    nextButton.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(ActionEvent e) {
        nextClicked(e);
      }
    });

    filenamesList.setToolTipText("List of available music files");
    listPanel.setLayout(new BorderLayout());
    menuFile.add(menuFileExit);
    menuHelp.add(menuHelpAbout);
    playerMenu.add(menuFile);
    playerMenu.add(menuHelp);
    this.setJMenuBar(playerMenu);
    refContentPane.add(listPanel, BorderLayout.CENTER);
    listPanel.add(filenamesList,  BorderLayout.CENTER);
    refContentPane.add(buttonPanel, BorderLayout.SOUTH);
    buttonPanel.add(prevButton, null);
    buttonPanel.add(playButton, null);
    buttonPanel.add(pauseButton, null);
    buttonPanel.add(stopButton, null);
    buttonPanel.add(nextButton, null);

  }


  private void logicInit() {
    Properties myProp = null;
    FileInputStream inFileStream = null ;
    searchDir = DEFAULT_DIR;
    try {
        inFileStream = new FileInputStream(new File(CONFIG_FILE_NAME));
        myProp = new Properties();
        myProp.load(inFileStream);
      } catch (Exception ex)
          { System.out.println("Config file access problem");  }
    if (myProp != null)  {
     searchDir = myProp.getProperty(MUSIC_DIR_PROPERTY,DEFAULT_DIR);
     }
     File tpDir = new File(searchDir);
     if (tpDir.isDirectory()) {
        String [] tpFiles = tpDir.list(new java.io.FilenameFilter() {

           public boolean accept(File indir, String inName) {
              String lcName = inName.toLowerCase();
              if (lcName.endsWith("ogg") || lcName.endsWith("mp3")
                || lcName.endsWith("wav"))
                return true;
              else
               return false;
           }

        });
        DefaultListModel dm = new DefaultListModel();
        curPlayListLength = tpFiles.length;
        for (int i=0; i<curPlayListLength; i++)
          dm.addElement(tpFiles[i]);
       filenamesList.setModel(dm);

     }
  }
  public void menuFileExitClicked(ActionEvent e) {
    pBase.endThread();
    System.exit(0);
  }

  public void menuHelpAboutClicked(ActionEvent e) {
    AboutBox dlg = new AboutBox(this);
    Dimension dlgSize = dlg.getPreferredSize();
    Dimension frmSize = getSize();
    Point loc = getLocation();
    dlg.setLocation((frmSize.width - dlgSize.width) / 2 + loc.x, (frmSize.height - dlgSize.height) / 2 + loc.y);
    dlg.setModal(true);
    dlg.show();
  }

  protected void processWindowEvent(WindowEvent e) {
    super.processWindowEvent(e);
    if (e.getID() == WindowEvent.WINDOW_CLOSING) {
      menuFileExitClicked(null);
    }
  }

  void playClicked(ActionEvent e) {
   String fileToPlay = (String) filenamesList.getSelectedValue();
   if (fileToPlay != null) {
        pBase.playFile(searchDir + System.getProperty("file.separator") + fileToPlay);
   }
  }
 void stopClicked(ActionEvent e) {
   pBase.stop();
  }
 void pauseClicked(ActionEvent e) {
   pBase.pause();
  }

  void prevClicked(ActionEvent e) {
    pBase.stop();
    filenamesList.setSelectedIndex(   filenamesList.getSelectedIndex() - 1);
    playClicked(e);

  }
  void nextClicked(ActionEvent e) {
    pBase.stop();
    filenamesList.setSelectedIndex(   (filenamesList.getSelectedIndex() + 1) % curPlayListLength);
   playClicked(e);
  }

  private void addButtonIconText(JButton inBut, String inTxt, String inIcon) {
    ImageIcon tpIcon = new ImageIcon(com.ddj.ddjplayer.DdjFrame.class.getResource(inIcon));
    inBut.setIcon(tpIcon);
    inBut.setHorizontalTextPosition(SwingConstants.CENTER);
    inBut.setVerticalTextPosition(SwingConstants.BOTTOM);
    inBut.setText(inTxt);
  }
}